"""Wrappers."""
from motor.core import AgnosticBaseProperties
from pymongo.collection import Collection
from pymongo.database import Database
from quart import abort
from motor import motor_asyncio


class AsyncIOMotorClient(motor_asyncio.AsyncIOMotorClient):
    """Wrapper for :class:`AsyncIOMotorClient.MongoClient`.

    Returns instances of Quart-Motor
    :class:`~quart_motor.wrappers.Database` instead of native motor
    :class:`~AsyncIOMotorDatabase` when accessed with dot notation.
    """

    def __getitem__(self, name):
        """__getitem__."""
        return AsyncIOMotorDatabase(self, name)


class AsyncIOMotorDatabase(motor_asyncio.AsyncIOMotorDatabase):
    """Wrapper for :class:`~AsyncIOMotorDatabase`.

    Returns instances of Quart-Motor
    :class:`~quart_motor.wrappers.AsyncIOMotorCollection` instead of native PyMongo
    :class:`~motor.motor_asyncio.AsyncIOMotorCollection` when accessed with dot notation.
    """

    def __init__(self, client, name, **kwargs):
        """__init__."""
        self._client = client
        delegate = kwargs.get("_delegate") or Database(client.delegate, name, **kwargs)

        super(AgnosticBaseProperties, self).__init__(delegate)

    def __getitem__(self, name):
        """__getitem__."""
        return AsyncIOMotorCollection(self, name)


class AsyncIOMotorCollection(motor_asyncio.AsyncIOMotorCollection):
    """Sub-class of Motor :class:`~AsyncIOMotorCollection` with helpers."""

    def __init__(
        self,
        database,
        name,
        codec_options=None,
        read_preference=None,
        write_concern=None,
        read_concern=None,
        _delegate=None,
    ):
        """__init__."""
        db_class = AsyncIOMotorDatabase

        if not isinstance(database, db_class):
            raise TypeError(
                "First argument to MotorCollection must be "
                "MotorDatabase, not %r" % database
            )

        delegate = _delegate or Collection(
            database.delegate,
            name,
            codec_options=codec_options,
            read_preference=read_preference,
            write_concern=write_concern,
            read_concern=read_concern,
        )

        super(AgnosticBaseProperties, self).__init__(delegate)
        self.database = database

    def __getitem__(self, name):
        """__getitem__."""
        return AsyncIOMotorCollection(
            self.database, self.name + "." + name, _delegate=self.delegate[name]
        )

    async def find_one_or_404(self, *args, **kwargs):
        """Find a single document or raise a 404.

        This is like :meth:`~AsyncIOMotorCollection.Collection.find_one`, but
        rather than returning ``None``, cause a 404 Not Found HTTP status
        on the request.
        .. code-block:: python
            @app.route("/user/<username>")
            def user_profile(username):
                user = mongo.db.users.find_one_or_404({"_id": username})
                return render_template("user.html",
                    user=user)
        """
        found = await self.find_one(*args, **kwargs)
        if found is None:
            abort(404)
        return found



import pytest
from quart import Quart
from quart.exceptions import NotFound

from quart_motor import Motor


class TestCollection:
    uri = f"mongodb://localhost:27017/test"

    @pytest.mark.asyncio
    async def test_find_one_or_404_notfound(self):
        app = Quart(__name__)
        mongo = Motor(app=app, uri=self.uri)
        await app.startup()
        await mongo.db.things.delete_many({})
        with pytest.raises(NotFound):
            await mongo.db.things.find_one_or_404({"_id": "thing"})

    @pytest.mark.asyncio
    async def test_find_one_or_404_found(self):
        app = Quart(__name__)
        mongo = Motor(app=app, uri=self.uri)
        await app.startup()
        await mongo.db.things.insert_one({"_id": "thing", "val": "foo"})
        thing = await mongo.db.things.find_one_or_404({"_id": "thing"})
        assert thing["val"] == "foo"
        await mongo.db.things.delete_many({})



"""Helpers."""
from bson import json_util, SON
from bson.errors import InvalidId
from bson.objectid import ObjectId
from quart import abort, json as quart_json
from six import iteritems, string_types
from werkzeug.routing import BaseConverter
import pymongo

__all__ = ["BSONObjectIdConverter", "JSONEncoder"]


if pymongo.version_tuple >= (3, 5, 0):
    from bson.json_util import RELAXED_JSON_OPTIONS
    DEFAULT_JSON_OPTIONS = RELAXED_JSON_OPTIONS
else:
    DEFAULT_JSON_OPTIONS = None


def _iteritems(obj):
    if hasattr(obj, "iteritems"):
        return obj.iteritems()
    elif hasattr(obj, "items"):
        return obj.items()
    else:
        raise TypeError("{!r} missing iteritems() and items()".format(obj))


class BSONObjectIdConverter(BaseConverter):
    """A simple converter for the RESTful URL routing system of Quart.

    .. code-block:: python
        @app.route("/<ObjectId:task_id>")
        def show_task(task_id):
            task = mongo.db.tasks.find_one_or_404(task_id)
            return render_template("task.html", task=task)
    Valid object ID strings are converted into
    :class:`~bson.objectid.ObjectId` objects; invalid strings result
    in a 404 error. The converter is automatically registered by the
    initialization of :class:`~quart_motor.Motor` with keyword
    :attr:`ObjectId`.
    The :class:`~quart_motor.helpers.BSONObjectIdConverter` is
    automatically installed on the :class:`~quart_motor.Motor`
    instance at creation time.
    """

    def to_python(self, value):
        """to_python."""
        try:
            return ObjectId(value)
        except InvalidId:
            raise abort(404)

    def to_url(self, value):
        """to_url."""
        return str(value)


class JSONEncoder(quart_json.JSONEncoder):
    """A JSON encoder that uses :mod:`bson.json_util` for MongoDB documents.

    .. code-block:: python
        @app.route("/cart/<ObjectId:cart_id>")
        def json_route(cart_id):
            results = mongo.db.carts.find({"_id": cart_id})
            return jsonify(results)
        # returns a Response with JSON body and application/json content-type:
        # '[{"count":12,"item":"egg"},{"count":1,"item":"apple"}]'
    Since this uses PyMongo's JSON tools, certain types may serialize
    differently than you expect. See :class:`~bson.json_util.JSONOptions`
    for details on the particular serialization that will be used.
    A :class:`~quart_motor.helpers.JSONEncoder` is automatically
    automatically installed on the :class:`~quart_motor.Motor`
    instance at creation time, using
    :const:`~bson.json_util.RELAXED_JSON_OPTIONS`. You can change the
    :class:`~bson.json_util.JSONOptions` in use by passing
    ``json_options`` to the :class:`~quart_motor.Motor`
    constructor.
    .. note::
        :class:`~bson.json_util.JSONOptions` is only supported as of
        PyMongo version 3.4. For older versions of PyMongo, you will
        have less control over the JSON format that results from calls
        to :func:`~Quart.json.jsonify`.
    .. versionadded:: 2.4.0
    """

    def __init__(self, json_options, *args, **kwargs):
        """__init__."""
        if json_options is None:
            json_options = DEFAULT_JSON_OPTIONS
        if json_options is not None:
            self._default_kwargs = {"json_options": json_options}
        else:
            self._default_kwargs = {}

        super(JSONEncoder, self).__init__(*args, **kwargs)

    def default(self, obj):
        """Serialize MongoDB object types using :mod:`bson.json_util`.

        Falls back to Quart's default JSON serialization for all other types.
        This may raise ``TypeError`` for object types not recognized.
        .. version-added:: 2.4.0
        """
        if hasattr(obj, "iteritems") or hasattr(obj, "items"):
            return SON((k, self.default(v)) for k, v in iteritems(obj))
        elif hasattr(obj, "__iter__") and not isinstance(obj, string_types):
            return [self.default(v) for v in obj]
        else:
            try:
                return json_util.default(obj, **self._default_kwargs)
            except TypeError:
                # PyMongo couldn't convert into a serializable object, and
                # the Flask default JSONEncoder won't; so we return the
                # object itself and let stdlib json handle it if possible
                return obj



"""
Quart-Motor
-------------
Async MongoDB support for Quart applications.
Quart-Motor is pip-installable:
    $ pip install Quart-Motor

Source code is hosted on `GitHub <https://github.com/marirs/quart-motor>`_.
Contributions are welcome!
"""

from setuptools import find_packages, setup

tests_require = [
    'coverage>=4.5',
    'coveralls>=1.8',
    'pytest>=5.2',
    'pytest-cov>=2.8',
    'pytest-pep8>=1.0',
    'pydocstyle>=4.0',
]

extras_require = {
    'tests': tests_require,
}

extras_require['all'] = [req for exts, reqs in extras_require.items()
                         for req in reqs]

setup_requires = [
    'pytest-runner>=5.2',
]

with open('README.md') as f:
    long_description = f.read()

setup(
    name="Quart-Motor",
    version='2.4.4',
    url="https://www.github.com/marirs/quart-motor/",
    download_url="https://www.github.com/marirs/quart-motor/tags",
    license="BSD",
    author="Sriram G",
    author_email="marirs@gmail.com",
    description="Motor support for Quart applications",
    long_description=long_description,
    long_description_content_type='text/markdown',
    zip_safe=False,
    packages=find_packages(),
    install_requires=[
        "PyMongo>=3.10",
        "Quart>=0.12.0",
        "motor>=2.1.0",
        "six",
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    python_requires='>=3.6',
)


import time

from quart import Quart
from motor.motor_asyncio import AsyncIOMotorDatabase
from pymongo.errors import InvalidURI

from quart_motor import Motor
import pytest


class CouldNotConnect(Exception):
    pass


class TestQuartMotor:
    uri = f"mongodb://localhost:27017/test"
    client_uri = f"mongodb://localhost:27017/"

    @pytest.mark.asyncio
    async def test__motor_client_connection(self):
        app = Quart(__name__)
        mongo = Motor(app=app, uri=self.client_uri)
        assert isinstance(mongo, Motor)

    @pytest.mark.asyncio
    async def test_motor_invalid_uri(self):
        app = Quart(__name__)
        with pytest.raises(InvalidURI):
            mongo = Motor(app=app, uri="http://localhost:27017/test")

    @pytest.mark.asyncio
    async def test_motor_database_success(self):
        app = Quart(__name__)
        mongo = Motor(app=app, uri=self.uri)
        await app.startup()
        assert isinstance(mongo.db, AsyncIOMotorDatabase)

    @pytest.mark.asyncio
    async def test_motor_database_failure(self):
        app = Quart(__name__)
        with pytest.raises(ValueError):
            mongo = Motor(app=app)

    @pytest.mark.asyncio
    async def test_multiple_motor_db_connections(self):
        app = Quart(__name__)
        db1 = Motor(app=app, uri=f"{self.client_uri}test1")
        db2 = Motor(app=app, uri=f"{self.client_uri}test2")
        await app.startup()
        assert isinstance(db1.db, AsyncIOMotorDatabase)
        assert isinstance(db2.db, AsyncIOMotorDatabase)

    @pytest.mark.asyncio
    async def test_motor_no_database_name_in_uri(self):
        app = Quart(__name__)
        mongo = Motor(app=app, uri=self.client_uri)
        await app.startup()
        assert mongo.db is None

    @pytest.mark.asyncio
    async def test_motor_custom_document_class(self):
        class CustomDict(dict):
            pass

        app = Quart(__name__)
        mongo = Motor(app=app, uri=self.uri, document_class=CustomDict)
        await app.startup()
        things = await mongo.db.things.find_one()
        assert things is None

        await mongo.db.things.insert_one({"_id": "thing", "val": "foo"})
        things = await mongo.db.things.find_one()
        await mongo.db.things.delete_many({})
        assert type(things) == CustomDict

    @pytest.mark.asyncio
    async def test_motor_doesnt_connect_by_default(self):
        app = Quart(__name__)
        mongo = Motor(app=app, uri=self.uri)
        await app.startup()

        with pytest.raises(CouldNotConnect):
            _wait_until_connected(mongo, timeout=0.2)


def _wait_until_connected(mongo, timeout=1.0):
    start = time.time()
    while time.time() < (start + timeout):
        if mongo.cx.nodes:
            return
        time.sleep(0.05)
    raise CouldNotConnect("could not prove mongodb connected in %r seconds" % timeout)


"""Quart-Motor.

Description: Motor for Quart
Modified from: https://github.com/dcrosta/flask-pymongo
Author: Sriram
"""
import asyncio

import pymongo
from functools import partial
from mimetypes import guess_type
import sys

from quart import abort, current_app, request
from gridfs import GridFS, NoFile
from pymongo import uri_parser
from werkzeug.wsgi import wrap_file

from quart_motor.helpers import BSONObjectIdConverter, JSONEncoder
from quart_motor.wrappers import AsyncIOMotorClient

__all__ = ("Motor", "ASCENDING", "DESCENDING")

PY2 = sys.version_info[0] == 2

# Python 3 compatibility
if PY2:
    text_type = (str, unicode)
    num_type = (int, long)
else:
    text_type = str
    num_type = int


DESCENDING = pymongo.DESCENDING
"""Descending sort order."""

ASCENDING = pymongo.ASCENDING
"""Ascending sort order."""


class Motor(object):
    """Manages MongoDB connections for your Quart app.

    Motor objects provide access to the MongoDB server via the :attr:`db`
    and :attr:`cx` attributes. You must either pass the :class:`~quart.Quart`
    app to the constructor, or call :meth:`init_app`.
    Motor accepts a MongoDB URI via the ``MONGO_URI`` QUART configuration
    variable, or as an argument to the constructor or ``init_app``. See
    :meth:`init_app` for more detail.
    """

    def __init__(self, app=None, uri=None, json_options=None, *args, **kwargs):
        """__init__."""
        self.cx = None
        self.db = None
        self._json_encoder = partial(JSONEncoder, json_options=json_options)

        if app is not None:
            self.init_app(app, uri, *args, **kwargs)

    def init_app(self, app, uri=None, *args, **kwargs):
        """Initialize this :class:`Motor` for use.

        Configure a :class:`~motor_async.AsyncIOMotorClient`
        in the following scenarios:
        1. If ``uri`` is not ``None``, pass the ``uri`` and any positional
           or keyword arguments to :class:`~motor_async.AsyncIOMotorClient`
        2. If ``uri`` is ``None``, and a Quart config variable named
           ``MONGO_URI`` exists, use that as the ``uri`` as above.
        The caller is responsible for ensuring that additional positional
        and keyword arguments result in a valid call.
        .. version-changed:: 2.2
           The ``uri`` is no longer required to contain a database name. If it
           does not, then the :attr:`db` attribute will be ``None``.
        .. version-changed:: 2.0
           Quart-Motor no longer accepts many of the configuration variables
           it did in previous versions. You must now use a MongoDB URI to
           configure Quart-Motor.
        """
        async def _before_serving():
            self.cx = AsyncIOMotorClient(*args, **kwargs)
            if database_name:
                self.db = self.cx[database_name]

        if uri is None:
            uri = app.config.get("MONGO_URI", None)
        if uri is not None:
            args = tuple([uri] + list(args))
        else:
            raise ValueError(
                "You must specify a URI or set the MONGO_URI Quart config variable",
            )

        parsed_uri = uri_parser.parse_uri(uri)
        database_name = parsed_uri["database"]

        # Try to delay connecting, in case the app is loaded before forking, per
        # http://api.mongodb.com/python/current/faq.html#is-pymongo-fork-safe
        kwargs.setdefault("connect", False)

        app.before_serving(_before_serving)

        app.url_map.converters["ObjectId"] = BSONObjectIdConverter
        app.json_encoder = self._json_encoder

    # view helpers
    def send_file(self, filename, base="fs", version=-1, cache_for=31536000):
        """Respond with a file from GridFS.

        Returns an instance of the :attr:`~quart.Quart.response_class`
        containing the named file, and implement conditional GET semantics
        (using :meth:`~werkzeug.wrappers.ETagResponseMixin.make_conditional`).
        .. code-block:: python
            @app.route("/uploads/<path:filename>")
            def get_upload(filename):
                return mongo.send_file(filename)
        :param str filename: the filename of the file to return
        :param str base: the base name of the GridFS collections to use
        :param bool version: if positive, return the Nth revision of the file
           identified by filename; if negative, return the Nth most recent
           revision. If no such version exists, return with HTTP status 404.
        :param int cache_for: number of seconds that browsers should be
           instructed to cache responses
        """
        if not isinstance(base, text_type):
            raise TypeError("'base' must be string or unicode")
        if not isinstance(version, num_type):
            raise TypeError("'version' must be an integer")
        if not isinstance(cache_for, num_type):
            raise TypeError("'cache_for' must be an integer")

        storage = GridFS(self.db, base)

        try:
            fileobj = storage.get_version(filename=filename, version=version)
        except NoFile:
            abort(404)

        # mostly copied from quart/helpers.py, with
        # modifications for GridFS
        data = wrap_file(request.environ, fileobj, buffer_size=1024 * 255)
        response = current_app.response_class(
            data,
            mimetype=fileobj.content_type,
            direct_passthrough=True,
        )
        response.content_length = fileobj.length
        response.last_modified = fileobj.upload_date
        response.set_etag(fileobj.md5)
        response.cache_control.max_age = cache_for
        response.cache_control.public = True
        response.make_conditional(request)
        return response

    def save_file(self, filename, fileobj, base="fs", content_type=None, **kwargs):
        """Save a file-like object to GridFS using the given filename.

        .. code-block:: python
            @app.route("/uploads/<path:filename>", methods=["POST"])
            def save_upload(filename):
                mongo.save_file(filename, request.files["file"])
                return redirect(url_for("get_upload", filename=filename))
        :param str filename: the filename of the file to return
        :param file fileobj: the file-like object to save
        :param str base: base the base name of the GridFS collections to use
        :param str content_type: the MIME content-type of the file. If
           ``None``, the content-type is guessed from the filename using
           :func:`~mimetypes.guess_type`
        :param kwargs: extra attributes to be stored in the file's document,
           passed directly to :meth:`gridfs.GridFS.put`
        """
        if not isinstance(base, text_type):
            raise TypeError("'base' must be string or unicode")
        if not (hasattr(fileobj, "read") and callable(fileobj.read)):
            raise TypeError("'fileobj' must have read() method")

        if content_type is None:
            content_type, _ = guess_type(filename)

        storage = GridFS(self.db, base)
        id = storage.put(fileobj, filename=filename, content_type=content_type, **kwargs)
        return id
